#include "sha512.c"
#include "fe_tobytes.c"
#include "fe_copy.c"
#include "fe_1.c"
#include "ge_tobytes.c"
#include "sign.c"
#include "fe_sq2.c"
#include "fe_cmov.c"
#include "ge_p1p1_to_p3.c"
#include "ge_p3_dbl.c"
#include "fe_mul.c"
#include "ge_p3_to_cached.c"
#include "fe_pow22523.c"
#include "ge_add.c"
#include "ge_sub.c"
#include "fe_add.c"
#include "ge_p2_0.c"
#include "ge_p3_to_p2.c"
#include "fe_invert.c"
#include "open.c"
#include "ge_msub.c"
#include "fe_0.c"
#include "fe_sub.c"
#include "fe_isnonzero.c"
#include "fe_isnegative.c"
#include "fe_neg.c"
#include "ge_frombytes.c"
#include "ge_p3_tobytes.c"
#include "ge_double_scalarmult.c"
#include "ge_p1p1_to_p2.c"
#include "fe_sq.c"
#include "ge_p3_0.c"
#include "ge_precomp_0.c"
#include "ge_scalarmult_base.c"
#include "sc_reduce.c"
#include "keypair.c"
#include "fe_frombytes.c"
#include "sc_muladd.c"
#include "ge_madd.c"
#include "ge_p2_dbl.c"
#include "crypto_verify.c"
#include "randombytes.c"
